{}
{                                                       }
{      Virtual Pascal v1.1                              }
{      Fast multiplex semaphore class                   }
{      }
{      Copyright (C) 1996 fPrint UK Ltd                 }
{                                                       }
{}

{$Delphi+,Frame-,LocInfo+,Optimise+,SmartLink+}

Unit Mutex;

Interface

uses
  os2def, os2base;

type
  tMutexSem = class
  public
    Next  : tMutexSem; // Next thread ID requesting ownership
    Owner : TID;       // Current semaphore owner; bit 31 = semaphore in use
    Ev    : hEv;       // Event semaphore for releasing
    Count : Longint;   // For recursive semaphore requests

    // Constructor; must be called before the class is used
    constructor create;
    // Request exclusive semaphore
    function request: boolean;
    // Release exclusive semaphore
    function release: boolean;
    // Query owner and request count for semaphore
    // This information is not guaranteed to be reliable
    procedure query(var Thread: Tid; var Cnt: Longint);
  private

  end;

Implementation

threadvar
  WakeUpSem: hEv;

const
  PostCount: Longint = 0;

// GetThreadSem creates a unique event semaphore per thread that needs
// it for MutexSem handling.  Only one semaphore is created per thread.
function GetThreadSem: hEv;
begin
  Result := WakeUpSem;
  if Result = 0 then
  begin
    DosCreateEventSem( nil, WakeUpSem, 0, False );
    Result := WakeUpSem;
  end;
end;

constructor tMutexSem.create;
begin
  Owner := 0;
  Next := nil;
  Count := 0;
  Ev := 0;
end;

function tMutexSem.Request; assembler {&frame-} {&uses none};
asm
            mov     eax,fs:[12]          // Get ^Thread Information Block
            push    0                                     // Event Sem: 0
            push    dword ptr [eax]                    // Owner : Longint
            push    eax                                 // Next : Pointer
            mov     ecx,Self+12                        // +12 b.o. Frame-
            push    dword ptr [ecx]                     // Class info ptr
@@testSem:
       lock bts     [ecx].Owner,31                     // Sem is now busy
            jnc     @@semFree
            push    ecx                               // Was already busy
            push    1        // There is no hurry since semaphore is busy
            call    DosSleep                // Go to sleep for a while...
            pop     eax
            pop     ecx
            jmp     @@testSem

@@semFree:
            mov     edx,[ecx].Owner                // Get semaphore owner
            btr     edx,31                              // Reset busy bit
            cmp     edx,[esp].Owner               // Owner = current TID?
            jne     @@notOur
            inc     [ecx].Count               // Yes, just increase count
       lock btr     [ecx].Owner,31                   // Release semaphore
            add     esp,20
            mov     al,1
            ret     4

@@notOur:
            mov     eax,esp
            xchg    eax,[ecx].Next
            test    edx,edx                                 // Owner = 0?
            jz      @@notBusy
            mov     [esp].Next,eax                       // Save ^nextTID
            mov     edx,dword ptr [esp].Owner                  // Our TID

            push    ecx
            call    GetThreadSem      // Get semaphore for current thread
            pop     ecx

            mov     [esp].Ev,eax                 // Save semaphore handle
       lock btr     [ecx].Owner,31                   // Release semaphore

            push    sem_indefinite_wait                   // Wait forever
            push    eax                      // Wait until it is our turn
            call    DosWaitEventSem           // Restore semahpore handle
       lock btr     [ecx].Owner,31                   // Release semaphore
            pop     edx
            pop     eax

            lea     eax,PostCount
            push    eax                           // Address of postcount
            push    edx                               // Semaphore handle
            call    DosResetEventSem               // Reset for next wait
            pop     edx                                 // Clean up stack
            pop     edx

            add     esp,16
            mov     al,1
            ret     4

@@notBusy:
            xchg    eax,[ecx].Next
            inc     edx
            mov     [ecx].Count,edx                  // Request count = 1
            pop     edx                                  // Skip classptr
            pop     eax                                  // Skip ^nextTID
            pop     eax                                      // Get owner
            pop     edx                                        // Skip Ev
       lock xchg    [ecx].Owner,eax         // Set owner&unlock semaphore
            mov     al,1
end;

function tMutexSem.Release; assembler {&frame-} {&uses none};
asm
@@testSem:
            mov     ecx,Self
       lock bts     [ecx].Owner,31              // Lock semaphore updates
            jnc     @@semFree
            push    1                                     // Already busy
            call    DosSleep                                   // Wait...
            pop     eax
            jmp     @@testSem                                // And retry

@@semFree:
            mov     eax,fs:[12]                               // Get ^TIB
            mov     eax,[eax]                              // GetThreadID
            bts     eax,31            // Set bit 31 in EAX for comparison
            cmp     eax,[ecx].Owner                      // Do we own it?
            je      @@isOur
       lock btr     [ecx].Owner,31         // No, so we cannot release it
            mov     al,0                                  // Return false
            ret     4

@@isOur:
            dec     [ecx].Count                     // Request count = 1?
            jz      @@scanChain
       lock btr     [ecx].Owner,31                            // No, done
            mov     al,1
            ret     4

@@scanChain:
            mov     edx,eax                        // Find next candidate
            mov     eax,ecx
            mov     ecx,[ecx].Next                            // ^nextTID
            test    ecx,ecx
            jz      @@1
            jnz     @@scanChain                   // Scan to end of chain
@@1:
            mov     ecx,Self
            cmp     eax,ecx                           // Are we the ones?
            je      @@onlyOwner                // Thread is only in chain

            mov     [edx].Next,0              // Remove thread from chain
            mov     [ecx].Count,1               // Set request count to 1
            mov     edx,[eax].Owner
            mov     eax,[eax].Ev       // Get semaphore handle to wake up
            bts     edx,31                                // Mark as busy
       lock xchg    [ecx].Owner,edx        // Make thread semaphore owner

            push    eax
            call    DosPostEventSem                  // Wake up new owner
            pop     eax
            mov     al,1
            ret     4

@@onlyOwner:
            xor     eax,eax               // Semaphore is no longer owned
       lock xchg    eax,[ecx].tMutexSem.Owner
            mov     al,1
end;

procedure tMutexSem.query(var Thread: Tid; var Cnt: Longint);
assembler {&frame-} {&uses none};
asm
            mov     eax,Self
            mov     ecx,[eax].Owner
            btr     ecx,31
            mov     edx,Thread
            mov     [edx],ecx
            mov     eax,[eax].count
            mov     edx,Cnt
            mov     [edx],eax
end;

end.

